@@ -21,19 +21,19 @@
 
 /**
  * SECTION:element-fluiddec
+ * @title: fluiddec
  * @see_also: timidity, wildmidi
  *
  * This element renders midi-events as audio streams using
  * <ulink url="http://fluidsynth.sourceforge.net//">Fluidsynth</ulink>.
  * It offers better sound quality compared to the timidity or wildmidi element.
  *
- * <refsect2>
- * <title>Example pipeline</title>
+ * ## Example pipeline
  * |[
  * gst-launch-1.0 filesrc location=song.mid ! midiparse ! fluiddec ! pulsesink
  * ]| This example pipeline will parse the midi and render to raw audio which is
  * played via pulseaudio.
- * </refsect2>
+ *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -114,25 +114,26 @@ G_DEFINE_TYPE (GstFluidDec, gst_fluid_dec, GST_TYPE_ELEMENT);
 
 /* fluid_synth log handler */
 static void
-gst_fluid_synth_error_log_function (int level, char *message, void *data)
+gst_fluid_synth_error_log_function (int level, const char *message, void *data)
 {
   GST_ERROR ("%s", message);
 }
 
 static void
-gst_fluid_synth_warning_log_function (int level, char *message, void *data)
+gst_fluid_synth_warning_log_function (int level, const char *message,
+    void *data)
 {
   GST_WARNING ("%s", message);
 }
 
 static void
-gst_fluid_synth_info_log_function (int level, char *message, void *data)
+gst_fluid_synth_info_log_function (int level, const char *message, void *data)
 {
   GST_INFO ("%s", message);
 }
 
 static void
-gst_fluid_synth_debug_log_function (int level, char *message, void *data)
+gst_fluid_synth_debug_log_function (int level, const char *message, void *data)
 {
   GST_DEBUG ("%s", message);
 }
@@ -177,10 +178,8 @@ gst_fluid_dec_class_init (GstFluidDecClass * klass)
           "Synth Polyphony", "The number of simultaneous voices", 1, 65535,
           DEFAULT_SYNTH_POLYPHONY, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
-  gst_element_class_add_pad_template (gstelement_class,
-      gst_static_pad_template_get (&src_factory));
-  gst_element_class_add_pad_template (gstelement_class,
-      gst_static_pad_template_get (&sink_factory));
+  gst_element_class_add_static_pad_template (gstelement_class, &src_factory);
+  gst_element_class_add_static_pad_template (gstelement_class, &sink_factory);
 
   gst_element_class_set_static_metadata (gstelement_class, "Fluidsynth",
       "Codec/Decoder/Audio",
@@ -189,14 +188,16 @@ gst_fluid_dec_class_init (GstFluidDecClass * klass)
   gstelement_class->change_state = gst_fluid_dec_change_state;
 
 #ifndef GST_DISABLE_GST_DEBUG
-  fluid_set_log_function (FLUID_PANIC, gst_fluid_synth_error_log_function,
-      NULL);
-  fluid_set_log_function (FLUID_ERR, gst_fluid_synth_warning_log_function,
-      NULL);
-  fluid_set_log_function (FLUID_WARN, gst_fluid_synth_warning_log_function,
-      NULL);
-  fluid_set_log_function (FLUID_INFO, gst_fluid_synth_info_log_function, NULL);
-  fluid_set_log_function (FLUID_DBG, gst_fluid_synth_debug_log_function, NULL);
+  fluid_set_log_function (FLUID_PANIC,
+      (fluid_log_function_t) gst_fluid_synth_error_log_function, NULL);
+  fluid_set_log_function (FLUID_ERR,
+      (fluid_log_function_t) gst_fluid_synth_warning_log_function, NULL);
+  fluid_set_log_function (FLUID_WARN,
+      (fluid_log_function_t) gst_fluid_synth_warning_log_function, NULL);
+  fluid_set_log_function (FLUID_INFO,
+      (fluid_log_function_t) gst_fluid_synth_info_log_function, NULL);
+  fluid_set_log_function (FLUID_DBG,
+      (fluid_log_function_t) gst_fluid_synth_debug_log_function, NULL);
 #else
   fluid_set_log_function (FLUID_PANIC, NULL, NULL);
   fluid_set_log_function (FLUID_ERR, NULL, NULL);
@@ -322,6 +323,7 @@ gst_fluid_dec_sink_event (GstPad * pad, GstObject * parent, GstEvent * event)
 
       res = gst_pad_push_event (fluiddec->srcpad, gst_event_new_caps (caps));
       gst_caps_unref (caps);
+      gst_event_unref (event);
       break;
     }
     case GST_EVENT_SEGMENT:
@@ -396,6 +398,9 @@ handle_buffer (GstFluidDec * fluiddec, GstBuffer * buffer)
 
   gst_buffer_map (buffer, &info, GST_MAP_READ);
 
+  if (info.size == 0)
+    goto done;
+
   event = info.data[0];
 
   switch (event & 0xf0) {
@@ -461,6 +466,9 @@ handle_buffer (GstFluidDec * fluiddec, GstBuffer * buffer)
       break;
     }
   }
+
+done:
+
   gst_buffer_unmap (buffer, &info);
 }
 
@@ -557,10 +565,14 @@ gst_fluid_dec_open (GstFluidDec * fluiddec)
           fluiddec->sf = fluid_synth_sfload (fluiddec->synth, filename, 1);
           if (fluiddec->sf != -1) {
             GST_DEBUG_OBJECT (fluiddec, "loaded soundfont file %s", filename);
+            g_free (filename);
+            g_dir_close (dir);
+            g_free (soundfont_path);
             goto done;
           }
           GST_DEBUG_OBJECT (fluiddec, "could not load soundfont file %s",
               filename);
+          g_free (filename);
         }
         g_dir_close (dir);
         g_free (soundfont_path);
