@@ -23,7 +23,7 @@
 
 #define VK_PROTOTYPES
 
-#include "vkconfig.h"
+#include <vkconfig.h>
 #include "vk_fwd.h"
 #include "vkmacros.h"
 
@@ -40,6 +40,18 @@
 #endif
 #endif
 
+#if GST_VULKAN_HAVE_WINDOW_COCOA
+#ifndef VK_USE_PLATFORM_MACOS_MVK
+#define VK_USE_PLATFORM_MACOS_MVK
+#endif
+#endif
+
+#if GST_VULKAN_HAVE_WINDOW_IOS
+#ifndef VK_USE_PLATFORM_IOS_MVK
+#define VK_USE_PLATFORM_IOS_MVK
+#endif
+#endif
+
 #include <vulkan/vulkan.h>
 
 #endif /* _VK_H_ */
