@@ -147,7 +147,7 @@ gst_wasapi_sink_class_init (GstWasapiSinkClass * klass)
 
   gst_element_class_add_static_pad_template (gstelement_class, &sink_template);
   gst_element_class_set_static_metadata (gstelement_class, "WasapiSrc",
-      "Sink/Audio",
+      "Sink/Audio/Hardware",
       "Stream audio to an audio capture device through WASAPI",
       "Nirbheek Chauhan <nirbheek@centricular.com>, "
       "Ole André Vadla Ravnås <ole.andre.ravnas@tandberg.com>");
@@ -211,6 +211,8 @@ gst_wasapi_sink_finalize (GObject * object)
   CoTaskMemFree (self->mix_format);
   self->mix_format = NULL;
 
+  CoUninitialize ();
+
   if (self->cached_caps != NULL) {
     gst_caps_unref (self->cached_caps);
     self->cached_caps = NULL;
@@ -220,8 +222,6 @@ gst_wasapi_sink_finalize (GObject * object)
   g_clear_pointer (&self->device_strid, g_free);
   self->mute = FALSE;
 
-  CoUninitialize ();
-
   G_OBJECT_CLASS (gst_wasapi_sink_parent_class)->finalize (object);
 }
 
@@ -404,7 +404,7 @@ gst_wasapi_sink_open (GstAudioSink * asink)
    * even if the old device was unplugged. We need to handle this somehow.
    * For example, perhaps we should automatically switch to the new device if
    * the default device is changed and a device isn't explicitly selected. */
-  if (!gst_wasapi_util_get_device_client (GST_ELEMENT (self), FALSE,
+  if (!gst_wasapi_util_get_device_client (GST_ELEMENT (self), eRender,
           self->role, self->device_strid, &device, &client)) {
     if (!self->device_strid)
       GST_ELEMENT_ERROR (self, RESOURCE, OPEN_WRITE, (NULL),
@@ -458,7 +458,7 @@ gst_wasapi_sink_get_can_frames (GstWasapiSink * self)
 
   /* Frames the card hasn't rendered yet */
   hr = IAudioClient_GetCurrentPadding (self->client, &n_frames_padding);
-  HR_FAILED_RET (hr, IAudioClient::GetCurrentPadding, -1);
+  HR_FAILED_ELEMENT_ERROR_RET (hr, IAudioClient::GetCurrentPadding, self, -1);
 
   GST_DEBUG_OBJECT (self, "%i unread frames (padding)", n_frames_padding);
 
@@ -480,12 +480,12 @@ gst_wasapi_sink_prepare (GstAudioSink * asink, GstAudioRingBufferSpec * spec)
   if (gst_wasapi_sink_can_audioclient3 (self)) {
     if (!gst_wasapi_util_initialize_audioclient3 (GST_ELEMENT (self), spec,
             (IAudioClient3 *) self->client, self->mix_format, self->low_latency,
-            &devicep_frames))
+            FALSE, &devicep_frames))
       goto beach;
   } else {
     if (!gst_wasapi_util_initialize_audioclient (GST_ELEMENT (self), spec,
             self->client, self->mix_format, self->sharemode, self->low_latency,
-            &devicep_frames))
+            FALSE, &devicep_frames))
       goto beach;
   }
 
@@ -557,6 +557,7 @@ gst_wasapi_sink_prepare (GstAudioSink * asink, GstAudioRingBufferSpec * spec)
 
   hr = IAudioClient_Start (self->client);
   HR_FAILED_GOTO (hr, IAudioClient::Start, beach);
+  self->client_needs_restart = FALSE;
 
   gst_audio_ring_buffer_set_channel_positions (GST_AUDIO_BASE_SINK
       (self)->ringbuffer, self->positions);
@@ -577,8 +578,6 @@ gst_wasapi_sink_unprepare (GstAudioSink * asink)
 {
   GstWasapiSink *self = GST_WASAPI_SINK (asink);
 
-  CoUninitialize ();
-
   if (self->client != NULL) {
     IAudioClient_Stop (self->client);
   }
@@ -588,6 +587,8 @@ gst_wasapi_sink_unprepare (GstAudioSink * asink)
     self->render_client = NULL;
   }
 
+  CoUninitialize ();
+
   return TRUE;
 }
 
@@ -597,59 +598,95 @@ gst_wasapi_sink_write (GstAudioSink * asink, gpointer data, guint length)
   GstWasapiSink *self = GST_WASAPI_SINK (asink);
   HRESULT hr;
   gint16 *dst = NULL;
-  guint pending = length;
+  DWORD dwWaitResult;
+  guint can_frames, have_frames, n_frames, write_len, written_len = 0;
 
   GST_OBJECT_LOCK (self);
   if (self->client_needs_restart) {
     hr = IAudioClient_Start (self->client);
-    HR_FAILED_AND (hr, IAudioClient::Start, GST_OBJECT_UNLOCK (self);
-        length = 0; goto beach);
+    HR_FAILED_ELEMENT_ERROR_AND (hr, IAudioClient::Start, self,
+        GST_OBJECT_UNLOCK (self); goto err);
     self->client_needs_restart = FALSE;
   }
   GST_OBJECT_UNLOCK (self);
 
-  while (pending > 0) {
-    DWORD dwWaitResult;
-    guint can_frames, have_frames, n_frames, write_len;
+  /* We have N frames to be written out */
+  have_frames = length / (self->mix_format->nBlockAlign);
 
+  if (self->sharemode == AUDCLNT_SHAREMODE_EXCLUSIVE) {
+    /* In exlusive mode we have to wait always */
     dwWaitResult = WaitForSingleObject (self->event_handle, INFINITE);
     if (dwWaitResult != WAIT_OBJECT_0) {
       GST_ERROR_OBJECT (self, "Error waiting for event handle: %x",
           (guint) dwWaitResult);
-      length -= pending;
-      goto beach;
+      goto err;
     }
 
-    /* We have N frames to be written out */
-    have_frames = pending / (self->mix_format->nBlockAlign);
-    /* We have can_frames space in the output buffer */
     can_frames = gst_wasapi_sink_get_can_frames (self);
-    /* We will write out these many frames, and this much length */
-    n_frames = MIN (can_frames, have_frames);
-    write_len = n_frames * self->mix_format->nBlockAlign;
+    if (can_frames < 0) {
+      GST_ERROR_OBJECT (self, "Error getting frames to write to");
+      goto err;
+    }
+    /* In exclusive mode we need to fill the whole buffer in one go or
+     * GetBuffer will error out */
+    if (can_frames != have_frames) {
+      GST_ERROR_OBJECT (self,
+          "Need at %i frames to write for exclusive mode, but got %i",
+          can_frames, have_frames);
+      goto err;
+    }
+  } else {
+    /* In shared mode we can write parts of the buffer, so only wait
+     * in case we can't write anything */
+    can_frames = gst_wasapi_sink_get_can_frames (self);
+    if (can_frames < 0) {
+      GST_ERROR_OBJECT (self, "Error getting frames to write to");
+      goto err;
+    }
 
-    GST_DEBUG_OBJECT (self, "total: %i, have_frames: %i (%i bytes), "
-        "can_frames: %i, will write: %i (%i bytes)", self->buffer_frame_count,
-        have_frames, pending, can_frames, n_frames, write_len);
-
-    hr = IAudioRenderClient_GetBuffer (self->render_client, n_frames,
-        (BYTE **) & dst);
-    HR_FAILED_AND (hr, IAudioRenderClient::GetBuffer, length = 0;
-        goto beach);
-
-    memcpy (dst, data, write_len);
-
-    hr = IAudioRenderClient_ReleaseBuffer (self->render_client, n_frames,
-        self->mute ? AUDCLNT_BUFFERFLAGS_SILENT : 0);
-    HR_FAILED_AND (hr, IAudioRenderClient::ReleaseBuffer, length = 0;
-        goto beach);
-
-    pending -= write_len;
+    if (can_frames == 0) {
+      dwWaitResult = WaitForSingleObject (self->event_handle, INFINITE);
+      if (dwWaitResult != WAIT_OBJECT_0) {
+        GST_ERROR_OBJECT (self, "Error waiting for event handle: %x",
+            (guint) dwWaitResult);
+        goto err;
+      }
+      can_frames = gst_wasapi_sink_get_can_frames (self);
+      if (can_frames < 0) {
+        GST_ERROR_OBJECT (self, "Error getting frames to write to");
+        goto err;
+      }
+    }
   }
 
-beach:
+  /* We will write out these many frames, and this much length */
+  n_frames = MIN (can_frames, have_frames);
+  write_len = n_frames * self->mix_format->nBlockAlign;
 
-  return length;
+  GST_DEBUG_OBJECT (self, "total: %i, have_frames: %i (%i bytes), "
+      "can_frames: %i, will write: %i (%i bytes)", self->buffer_frame_count,
+      have_frames, length, can_frames, n_frames, write_len);
+
+  hr = IAudioRenderClient_GetBuffer (self->render_client, n_frames,
+      (BYTE **) & dst);
+  HR_FAILED_ELEMENT_ERROR_AND (hr, IAudioRenderClient::GetBuffer, self,
+      goto err);
+
+  memcpy (dst, data, write_len);
+
+  hr = IAudioRenderClient_ReleaseBuffer (self->render_client, n_frames,
+      self->mute ? AUDCLNT_BUFFERFLAGS_SILENT : 0);
+  HR_FAILED_ELEMENT_ERROR_AND (hr, IAudioRenderClient::ReleaseBuffer, self,
+      goto err);
+
+  written_len = write_len;
+
+out:
+  return written_len;
+
+err:
+  written_len = -1;
+  goto out;
 }
 
 static guint
