@@ -49,8 +49,6 @@ struct _GstWasapiSrc
   /* Client was reset, so it needs to be started again */
   gboolean client_needs_restart;
 
-  /* Actual size of the allocated buffer */
-  guint buffer_frame_count;
   /* The mix format that wasapi prefers in shared mode */
   WAVEFORMATEX *mix_format;
   /* The probed caps that we can accept */
@@ -63,6 +61,7 @@ struct _GstWasapiSrc
   /* properties */
   gint role;
   gint sharemode;
+  gboolean loopback;
   gboolean low_latency;
   gboolean try_audioclient3;
   wchar_t *device_strid;
